import setuptools

setuptools.setup(
    name='nasbench_pytorch',
    version='1.0',
    license='Apache License 2.0',
    author='Romulus Hong, Gabriela Suchopárová',
    packages=setuptools.find_packages()
)
